#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#define GENERIC_ERROR_HELPER(cond, errCode, msg)     do {           \
        if (cond) {                                                 \
            fprintf(stderr, "%s: %s\n", msg, strerror(errCode));    \
            exit(EXIT_FAILURE);                                     \
        }                                                           \
    } while(0)

#define ERROR_HELPER(ret, msg)          GENERIC_ERROR_HELPER((ret < 0), errno, msg)
#define PTHREAD_ERROR_HELPER(ret, msg)  GENERIC_ERROR_HELPER((ret != 0), ret, msg)

#define BUFFER_SIZE     1024
#define CLOSE_COMMAND   "BYE"

#define FIFO_ACCEPT_SUFFIX  "_accept"
#define FIFO_CONNECT_SUFFIX "_connect"

char listen_fifo_name[128];
char accept_fifo_name[128];

int shouldStop = 0;

void* receiveMessage(void* arg) {
    int recv_fifo = (int)arg;

    char* close_command = CLOSE_COMMAND;
    size_t close_command_len = strlen(close_command);

    /* select() uses sets of descriptors and a timeval interval. The
     * methods returns when either an event occurs on a descriptor in
     * the sets during the given interval, or when that time elapses.
     *
     * The first argument for select is the maximum descriptor among
     * those in the sets plus one. Note also that both the sets and
     * the timeval argument are modified by the call, so you should
     * reinitialize them across multiple invocations.
     *
     * On success, select() returns the number of descriptors in the
     * given sets for which data may be available, or 0 if the timeout
     * expires before any event occurs. */
    struct timeval timeout;
    fd_set read_descriptors;
    int nfds = recv_fifo + 1;

    char buf[BUFFER_SIZE];

    while (!shouldStop) {
        int ret;

        // check every 1.5 seconds (why not longer?)
        timeout.tv_sec  = 1;
        timeout.tv_usec = 500000;

        FD_ZERO(&read_descriptors);
        FD_SET(recv_fifo, &read_descriptors);

        /** perform select() **/
        ret = select(nfds, &read_descriptors, NULL, NULL, &timeout);

        if (ret == -1 && errno == EINTR) continue;
        ERROR_HELPER(ret, "Unable to select()");

        if (ret == 0) continue; // timeout expired

        // ret is 1: read available data!

        /** COMPLETE CODE HERE
         *
         * Read from FIFO recv_fifo: store data in buf, and write the
         * number of bytes read in the variable ret. */


        // I'm assuming ret is now >= 0
        if (ret == 0) {
            fprintf(stderr, "[WARNING] Endpoint closed the FIFO. Exiting...\n");
            shouldStop = 1;
        } else if (ret == close_command_len && !memcmp(buf, close_command, close_command_len)) {
            fprintf(stderr, "Chat session terminated from endpoint. Please press ENTER to exit.\n");
            shouldStop = 1;
        } else {
            // print received message
            buf[ret] = '\0';
            printf("==> %s\n", buf);
        }
    }

    pthread_exit(NULL);
}

void* sendMessage(void* arg) {
    int send_fifo = (int)arg;

    char* close_command = CLOSE_COMMAND;
    size_t close_command_len = strlen(close_command);

    char buf[BUFFER_SIZE];

    while (!shouldStop) {
        int ret;

        /* Read a line from stdin: fgets() reads up to sizeof(buf)-1
         * bytes and on success returns the first argument passed. */
        if (fgets(buf, sizeof(buf), stdin) != (char*)buf) {
            fprintf(stderr, "Error while reading from stdin, exiting...\n");
            exit(EXIT_FAILURE);
        }

        if (shouldStop) break; // the endpoint might have closed the connection

        size_t msg_len = strlen(buf);
        --msg_len; // remove '\n' from the end of the message

        /** COMPLETE CODE HERE
         *
         * Write first msg_len bytes in buf to the FIFO send_fifo. */

        // After a BYE command we should update shouldStop
        if (msg_len == close_command_len && !memcmp(buf, close_command, close_command_len)) {
            shouldStop = 1;
            fprintf(stderr, "Chat session terminated.\n");
        }
    }

    pthread_exit(NULL);
}

void chatSession(int send_fifo, int recv_fifo) {
    int ret;

    fprintf(stderr, "Chat session started! Send %s to close it.\n", CLOSE_COMMAND);

    pthread_t chat_threads[2];

    ret = pthread_create(&chat_threads[0], NULL, receiveMessage, (void*)recv_fifo);
    PTHREAD_ERROR_HELPER(ret, "Cannot create thread for receiving messages");

    ret = pthread_create(&chat_threads[1], NULL, sendMessage, (void*)send_fifo);
    PTHREAD_ERROR_HELPER(ret, "Cannot create thread for sending messages");

    // wait for termination
    ret = pthread_join(chat_threads[0], NULL);
    PTHREAD_ERROR_HELPER(ret, "Cannot join on thread for receiving messages");

    ret = pthread_join(chat_threads[1], NULL);
    PTHREAD_ERROR_HELPER(ret, "Cannot join on thread for sending messages");

    /** COMPLETE CODE HERE
     *
     * Chat session terminated: close FIFO descriptors. */
}

// executed when user specifies a "connect" command
inline void connectOnFIFO() {
    int recv_fifo, send_fifo;

    /** COMPLETE CODE HERE
     *
     * Open FIFOs listen_fifo_name and accept_fifo_name. Store the
     * descriptor for the FIFO you will write to in send_fifo, and the
     * descriptor for the FIFO you will read from in recv_fifo.
     *
     * Hint: one of the two peers should create the two FIFOs, which
     *       should be destroyed when a chat session is completed.
     **/

    // start a chat session
    chatSession(send_fifo, recv_fifo);
}

// executed when user specifies an "accept" command
inline void listenOnFIFO() {
    int recv_fifo, send_fifo;

    /** COMPLETE CODE HERE
     *
     * Open FIFOs listen_fifo_name and accept_fifo_name. Store the
     * descriptor for the FIFO you will write to in send_fifo, and the
     * descriptor for the FIFO you will read from in recv_fifo.
     *
     * Hint: one of the two peers should create the two FIFOs, which
     *       should be destroyed when a chat session is completed.
     **/

    // start a chat session
    chatSession(send_fifo, recv_fifo);
}

void syntaxError(char* prog_name) {
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, "       %s accept <FIFO_prefix>\n", prog_name);
    fprintf(stderr, "  OR:\n");
    fprintf(stderr, "       %s connect <FIFO_prefix>\n", prog_name);
    exit(EXIT_FAILURE);
}

int main(int argc, char* argv[]) {
    if (argc == 3) {
        // the user specifies a prefix for the names of the two FIFOs we will use
        char* fifo_prefix = argv[2];
        if (!strcmp(argv[1], "accept")) {
            /* accept module will read from Listen FIFO and write to Accept FIFO */
            sprintf(listen_fifo_name, "%s%s", fifo_prefix, FIFO_ACCEPT_SUFFIX);
            sprintf(accept_fifo_name, "%s%s", fifo_prefix, FIFO_CONNECT_SUFFIX);

            listenOnFIFO();
        } else if (!strcmp(argv[1], "connect")) {
            /* connect module will read from Accept FIFO and write to Listen FIFO */
            sprintf(listen_fifo_name, "%s%s", fifo_prefix, FIFO_ACCEPT_SUFFIX);
            sprintf(accept_fifo_name, "%s%s", fifo_prefix, FIFO_CONNECT_SUFFIX);

            connectOnFIFO();
        } else {
            syntaxError(argv[0]);
        }
    } else {
        syntaxError(argv[0]);
    }
    exit(EXIT_SUCCESS);
}
